from __future__ import generator_stop

import ast
import operator
import timeit as real_timeit
import unittest
from functools import lru_cache
from inspect import signature
from io import StringIO
from itertools import chain
from pprint import pprint
from textwrap import dedent

import wrapt
from littleutils import only

from nbsafety.sorcery.core import spell, node_names, node_name

_NO_DEFAULT = object()


@spell
def assigned_names(frame_info):
    """
    Instead of:

        foo = func('foo')
        bar = func('bar')

    write:

        foo, bar = map(func, assigned_names())

    or:

        foo, bar = [func(name) for name in assigned_names()]

    Instead of:

        class Thing(Enum):
            foo = 'foo'
            bar = 'bar'

    write:

        class Thing(Enum):
            foo, bar = assigned_names()

    More generally, this function returns a tuple of strings representing the names being assigned to.

    The result can be assigned to any combination of either:

      - plain variables,
      - attributes, or
      - subscripts (square bracket access) with string literal keys

    So the following:

        spam, x.foo, y['bar'] = assigned_names()

    is equivalent to:

        spam = 'spam'
        x.foo = 'foo'
        y['bar'] = 'bar'

    Any expression is allowed to the left of the attribute/subscript.

    Only simple tuple unpacking is allowed:

      - no nesting, e.g.  (a, b), c = ...
      - no stars, e.g.    a, *b = ...
      - no chains, e.g.   a, b = c = ...
      - no assignment to a single name without unpacking, e.g.  a = ...
    """
    return frame_info.assigned_names(allow_one=True)[0]


@spell
def unpack_keys(frame_info, x, default=_NO_DEFAULT, prefix=None, swapcase=False):
    """
    Instead of:

        foo = d['foo']
        bar = d['bar']

    write:

        foo, bar = unpack_keys(d)

    Instead of:

        foo = d.get('foo', 0)
        bar = d.get('bar', 0)

    write:

        foo, bar = unpack_keys(d, default=0)

    Instead of:

        foo = d['data_foo']
        bar = d['data_bar']

    write:

        foo, bar = unpack_keys(d, prefix='data_')

    Instead of:

        foo = d['FOO']
        bar = d['BAR']

    write:

        foo, bar = unpack_keys(d, swapcase=True)

    and similarly, instead of:

        FOO = d['foo']
        BAR = d['bar']

    write:

        FOO, BAR = unpack_keys(d, swapcase=True)

    Note that swapcase is not applied to the prefix, so for example you should write:

        env = dict(DATABASE_USERNAME='me',
                   DATABASE_PASSWORD='secret')
        username, password = unpack_keys(env, prefix='DATABASE_', swapcase=True)

    The rules of the assigned_names spell apply.

    This can be seamlessly used in for loops, even inside comprehensions, e.g.

        for foo, bar in unpack_keys(list_of_dicts):
            ...

    If there are multiple assignment targets in the statement, e.g. if you have
    a nested list comprehension, the target nearest to the function call will
    determine the keys. For example, the keys 'foo' and 'bar' will be extracted in:

        [[foo + bar + y for foo, bar in unpack_keys(x)]
         for x, y in z]

    Like assigned_names, the unpack call can be part of a bigger expression,
    and the assignment will still be found. So for example instead of:

        foo = int(d['foo'])
        bar = int(d['bar'])

    you can write:

        foo, bar = map(int, unpack_keys(d))

    or:

        foo, bar = [int(v) for v in unpack_keys(d)]

    The second version works because the spell looks for multiple names being assigned to,
    so it doesn't just unpack 'v'.

    """

    if default is _NO_DEFAULT:
        getter = operator.getitem
    else:
        # Essentially dict.get, without relying on that method existing
        def getter(d, name):
            try:
                return d[name]
            except KeyError:
                return default

    return _unpack(frame_info, x, getter, prefix, swapcase)


@spell
def unpack_attrs(frame_info, x, default=_NO_DEFAULT, prefix=None, swapcase=False):
    """
    This is similar to unpack_keys, but for attributes.

    Instead of:

        foo = x.foo
        bar = x.bar

    write:

        foo, bar = unpack_attrs(x)
    """

    if default is _NO_DEFAULT:
        getter = getattr
    else:
        def getter(d, name):
            return getattr(d, name, default)

    return _unpack(frame_info, x, getter, prefix, swapcase)


def _unpack(frame_info, x, getter, prefix, swapcase):
    names, node = frame_info.assigned_names(allow_loops=True)

    def fix_name(n):
        if swapcase:
            n = n.swapcase()
        if prefix:
            n = prefix + n
        return n

    if isinstance(node, ast.Assign):
        return [getter(x, fix_name(name)) for name in names]
    else:  # for loop
        return ([getter(d, fix_name(name)) for name in names]
                for d in x)


@spell
def args_with_source(frame_info, *args):
    """
    Returns a list of pairs of:
        - the source code of the argument
        - the value of the argument
    for each argument.

    For example:

        args_with_source(foo(), 1+2)

    is the same as:

        [
            ("foo()", foo()),
            ("1+2", 3)
        ]
    """
    return [
        (frame_info.get_source(arg), value)
        for arg, value in zip(frame_info.call.args, args)
    ]


@spell
def dict_of(frame_info, *args, **kwargs):
    """
    Instead of:

        {'foo': foo, 'bar': bar, 'spam': thing()}

    or:

        dict(foo=foo, bar=bar, spam=thing())

    write:

        dict_of(foo, bar, spam=thing())

    In other words, returns a dictionary with an item for each argument,
    where positional arguments use their names as keys,
    and keyword arguments do the same as in the usual dict constructor.

    The positional arguments can be any of:

      - plain variables,
      - attributes, or
      - subscripts (square bracket access) with string literal keys

    So the following:

        dict_of(spam, x.foo, y['bar'])

    is equivalent to:

        dict(spam=spam, foo=x.foo, bar=y['bar'])

    *args are not allowed.

    To give your own functions the ability to turn positional argments into
    keyword arguments, use the decorator magic_kwargs.

    """

    result = {
        node_name(arg): value
        for arg, value in zip(frame_info.call.args[-len(args):], args)
    }
    result.update(kwargs)
    return result


@spell
def print_args(frame_info, *args, file=None):
    """
    For each argument, prints the source code of that argument
    and its value. Returns the first argument.
    """
    for source, arg in args_with_source.at(frame_info)(*args):
        print(source + ' =', file=file)
        pprint(arg, stream=file)
        print(file=file)
    return args and args[0]


@spell
def call_with_name(frame_info, func):
    """
    Given:
    
        class C:
            def generic(self, method_name, *args, **kwargs):
                ...
                
    Inside the class definition, instead of:

            def foo(self, x, y):
                return self.generic('foo', x, y)
            
            def bar(self, z):
                return self.generic('bar', z)
    
    write:
    
            foo, bar = call_with_name(generic)

    This only works for methods inside classes, not free functions.
    """
    def make_func(name):
        return lambda self, *args, **kwargs: func(self, name, *args, **kwargs)

    return [
        make_func(name)
        for name in frame_info.assigned_names()[0]
    ]


@spell
def delegate_to_attr(frame_info, attr_name):
    """
    This is a special case of the use case fulfilled by call_with_name.
    
    Given:
    
        class Wrapper:
            def __init__(self, thing):
                self.thing = thing
            
    Inside the class definition, instead of:
    
            def foo(self, x, y):
                return self.thing.foo(x, y)
            
            def bar(self, z):
                return self.thing.bar(z)

    Write:
    
            foo, bar = delegate_to_attr('thing')

    Specifically, this will make:

        Wrapper().foo

    equivalent to:

        Wrapper().thing.foo
    """
    def make_func(name):
        return property(lambda self: getattr(getattr(self, attr_name), name))

    return [
        make_func(name)
        for name in frame_info.assigned_names()[0]
    ]


class _Nothing(object):
    def __init__(self, count):
        self.__count = count

    def __getattribute__(self, item):
        if item == '_Nothing__count':
            return object.__getattribute__(self, item)
        return _Nothing.__op(self)

    def __op(self, *_args, **_kwargs):
        self.__count -= 1
        if self.__count == 0:
            return None

        return self

    __getitem__ = __call__ = __op


@spell
def maybe(frame_info, x):
    """
    Instead of:

        None if foo is None else foo.bar()

    write:

        maybe(foo).bar()

    Specifically, if foo is not None, then maybe(foo) is just foo.

    If foo is None, then any sequence of attributes, subscripts, or
    calls immediately to the right of maybe(foo) is ignored, and
    the final result is None. So maybe(foo)[0].x.y.bar() is None,
    while func(maybe(foo)[0].x.y.bar()) is func(None) because enclosing
    expressions are not affected.
    """
    if x is not None:
        return x

    node = frame_info.call
    count = 0
    while True:
        parent = node.parent
        if not (isinstance(parent, ast.Attribute) or
                isinstance(parent, ast.Call) and parent.func is node or
                isinstance(parent, ast.Subscript) and parent.value is node):
            break
        count += 1
        node = parent

    if count == 0:
        return x

    return _Nothing(count)


@spell
def select_from(frame_info, sql, params=(), cursor=None, where=None):
    """
    Instead of:

        cursor.execute('''
            SELECT foo, bar
            FROM my_table
            WHERE spam = ?
              AND thing = ?
            ''', [spam, thing])

        for foo, bar in cursor:
            ...

    write:

        for foo, bar in select_from('my_table', where=[spam, thing]):
            ...

    Specifically:
        - the assigned names (similar to the assigned_names and unpack_keys spells)
            are placed in the SELECT clause
        - the first argument (usually just a table name but can be any SQL)
            goes after the FROM
        - if the where argument is supplied, it must be a list or tuple literal of values
            which are supplied as query parameters and whose names are used in a
            WHERE clause using the = and AND operators.
            If you use this argument, don't put a WHERE clause in the sql argument and
            don't supply params
        - a cursor object is automatically pulled from the calling frame, but if this
            doesn't work you can supply one with the cursor keyword argument
        - the params argument can be supplied for more custom cases than the where
            argument provides.
        - if this is used in a loop or list comprehension, all rows in the result
            will be iterated over.
            If it is used in an assignment statement, one row will be returned.
        - If there are multiple names being assigned (i.e. multiple columns being selected)
            then the row will be returned and thus unpacked. If there is only one name,
            it will automatically be unpacked so you don't have to add [0].

    This spell is much more a fun rough idea than the others. It is expected that there
    are many use cases it will not fit into nicely.
    """
    if cursor is None:
        frame = frame_info.frame
        cursor = only(c for c in chain(frame.f_locals.values(),
                                       frame.f_globals.values())
                      if 'cursor' in str(type(c).__mro__).lower() and
                      callable(getattr(c, 'execute', None)))
    names, node = frame_info.assigned_names(allow_one=True, allow_loops=True)
    sql = 'SELECT %s FROM %s' % (', '.join(names), sql)

    if where:
        where_arg = only(kw.value for kw in frame_info.call.keywords
                         if kw.arg == 'where')
        where_names = node_names(where_arg)
        assert len(where_names) == len(where)
        sql += ' WHERE ' + ' AND '.join('%s = ?' % name for name in where_names)
        assert params == ()
        params = where

    cursor.execute(sql, params)

    def unpack(row):
        if len(row) == 1:
            return row[0]
        else:
            return row

    if isinstance(node, ast.Assign):
        return unpack(cursor.fetchone())
    else:
        def vals():
            for row in cursor:
                yield unpack(row)

        return vals()


def magic_kwargs(func):
    """
    Applying this decorator allows a function to interpret positional
    arguments as keyword arguments, using the name of the positional argument
    as the keyword. For example, given:

        @magic_kwargs
        def func(*, foo, bar, spam):

    or

        @magic_kwargs
        def func(**kwargs):

    then instead of:

        func(foo=foo, bar=bar, spam=thing)

    you can just write:

        func(foo, bar, spam=thing)

    Without the @magic_kwargs, the closest magical alternative would be:

        func(**dict_of(foo, bar, spam=thing))

    The function is not allowed to have optional positional parameters, e.g.
    `def func(x=1)`, or *args.
    """

    args_count = 0
    for param in signature(func).parameters.values():
        if (param.kind == param.VAR_POSITIONAL or
                param.kind == param.POSITIONAL_OR_KEYWORD and
                param.default != param.empty):
            raise TypeError(
                'The type of the parameter %s is not allowed with @magic_kwargs'
                % param.name)
        if param.kind == param.POSITIONAL_OR_KEYWORD:
            args_count += 1

    @wrapt.decorator
    def wrapper(wrapped, instance, args, kwargs):
        frame_info, *args = args
        count = args_count - (instance is not None)  # account for self argument
        normal_args = args[:count]
        magic_args = args[count:]
        full_kwargs = dict_of.at(frame_info)(*magic_args, **kwargs)
        return wrapped(*normal_args, **full_kwargs)

    return spell(wrapper(func))


@spell
def switch(frame_info, val, _cases, *, default=_NO_DEFAULT):
    """
    Instead of:

        if val == 1:
            x = 1
        elif val == 2 or val == bar():
            x = spam()
        elif val == dangerous_function():
            x = spam() * 2
        else:
            x = -1

    write:

        x = switch(val, lambda: {
            1: 1,
            {{ 2, bar() }}: spam(),
            dangerous_function(): spam() * 2
        }, default=-1)

    This really will behave like the if/elif chain above. The dictionary is just
    some nice syntax, but no dictionary is ever actually created. The keys
    are evaluated only as needed, in order, and only the matching value is evaluated.
    The keys are not hashed, only compared for equality, so non-hashable keys like lists
    are allowed.

    If the default is not specified and no matching value is found, a KeyError is raised.

    Note that `if val == 2 or val == bar()` is translated to `{{ 2, bar() }}`.
    This is to allow emulating multiple case clauses for the same block as in
    the switch construct in other languages. The double braces {{}} create a value
    that's impossible to evaluate normally (a set containing a set) so that it's clear
    we don't simply want to check `val == {{ 2, bar() }}`, whereas `{2, bar()}` would be
    evaluated and checked normally.
    As always, the contents are lazily evaluated and compared in order.

    The keys and values are evaluated with the compiler statement
    `from __future__ import generator_stop` in effect (which you should really be
    considering using anyway if you're using Python < 3.7).

    """

    frame = frame_info.frame
    switcher = _switcher(frame_info.call.args[1], frame.f_code)

    def ev(k):
        return eval(k, frame.f_globals, frame.f_locals)

    def check(k):
        return ev(k) == val

    for key_code, value_code in switcher:
        if isinstance(key_code, tuple):
            test = any(map(check, key_code))
        else:
            test = check(key_code)
        if test:
            return ev(value_code)

    if default is _NO_DEFAULT:
        raise KeyError(val)
    else:
        return default


@lru_cache()
def _switcher(cases, f_code):
    if not (isinstance(cases, ast.Lambda) and
            isinstance(cases.body, ast.Dict)):
        raise TypeError('The second argument to switch must be a lambda with no arguments '
                        'that returns a dictionary literal')

    def comp(node):
        return compile(ast.Expression(node),
                       filename=f_code.co_filename,
                       mode='eval')

    result = []
    for key, value in zip(cases.body.keys,
                          cases.body.values):

        if (isinstance(key, ast.Set) and
                isinstance(key.elts[0], ast.Set)):
            key_code = tuple(comp(k) for k in key.elts[0].elts)
        else:
            key_code = comp(key)

        result.append((key_code, comp(value)))
    return result


def _raise(e):
    # for tests
    raise e


class TimerWithExc(real_timeit.Timer):
    def timeit(self, *args, **kwargs):
        try:
            return super().timeit(*args, **kwargs)
        except:
            # Sets up linecache for future tracebacks
            self.print_exc(StringIO())
            raise


@spell
def timeit(frame_info, repeat=5):
    """
    This function is for writing quick scripts for comparing the speeds
    of two snippets of code that do the same thing. It's a nicer interface
    to the standard timeit module that doesn't require putting your code in strings, so you can
    use your IDE features, while still using the standard timeit for accuracy.
    
    Instead of
    
        import timeit
    
        nums = [3, 1, 2]
        setup = 'from __main__ import nums'
        
        print(timeit.repeat('min(nums)', setup))
        print(timeit.repeat('sorted(nums)[0]', setup))
    
    write:
    
        import sorcery
    
        nums = [3, 1, 2]
        
        if sorcery.timeit():
            result = min(nums)
        else:
            result = sorted(nums)[0]
            
    The if statement is just syntax for denoting the two blocks of code
    being tested. Some other nice features of this function over the standard
    timeit:
    
    - Automatically determines a high enough 'number' argument.
    - Asserts that any variable named 'result' is equal in both snippets,
      for correctness testing. The variable should be present in both or
      neither snippets.
    - Nice formatting of results for easy comparison, including best times
    - Source lines shown in tracebacks
    
    The spell must be called at the top level of a module, not inside
    another function definition.
    """
    
    globs = frame_info.frame.f_globals
    if globs is not frame_info.frame.f_locals:
        _raise(ValueError('Must execute in global scope'))

    setup = 'from %s import %s\n' % (
        globs['__name__'],
        ', '.join(globs.keys()),
    )
    if_stmt = frame_info.call.parent
    stmts = [
        dedent('\n'.join(map(frame_info.get_source, lines)))
        for lines in [if_stmt.body, if_stmt.orelse]
    ]

    timers = [
        TimerWithExc(stmt, setup)
        for stmt in stmts
    ]

    # Check for exceptions
    for timer in timers:
        timer.timeit(1)

    # Compare results
    def get_result(stmt):
        ns = {}
        exec(setup + stmt, ns)
        return ns.get('result')

    unittest.TestCase('__init__').assertEqual(
        *map(get_result, stmts),
        '\n=====\nThe two methods yielded different results!'
    )

    # determine number so that 1 <= total time < 3
    number = 1
    for i in range(22):
        number = 3 ** i
        if timers[0].timeit(number) >= 1:
            break

    print('Number of trials:', number)
    print()

    def print_time(idx, el):
        print('Method {}: {:.3f}'.format(
            idx + 1, el))

    times = [[] for _ in timers]
    for _ in range(repeat):
        for i, timer in enumerate(timers):
            elapsed = timer.timeit(number)
            print_time(i, elapsed)
            times[i].append(elapsed)
        print()

    print('Best times:')
    print('-----------')
    for i, elapsed_list in enumerate(times):
        print_time(i, min(elapsed_list))
    



import ast
import sys
from functools import lru_cache, partial
from types import CodeType
from typing import Set, Tuple

from executing import only, Source


class FrameInfo(object):
    """
    Contains metadata about where a spell is being called.
    An instance of this is passed as the first argument to any spell.
    Users should not instantiate this class themselves.

    There are two essential attributes:

    - frame: the execution frame in which the spell is being called
    - call: the ast.Call node where the spell is being called
        See https://greentreesnakes.readthedocs.io/en/latest/nodes.html
        to learn how to navigate the AST
    """

    def __init__(self, executing):
        self.frame = executing.frame
        self.executing = executing
        self.call = executing.node

    def assigned_names(self, *,
                       allow_one: bool = False,
                       allow_loops: bool = False
                       ) -> Tuple[Tuple[str, ...], ast.AST]:
        """
        Calls the function assigned_names for this instance's Call node.
        """
        return assigned_names(self.call,
                              allow_one=allow_one,
                              allow_loops=allow_loops)

    def get_source(self, node: ast.AST) -> str:
        """
        Returns a string containing the source code of an AST node in the
        same file as this call.
        """
        return self.executing.source.asttokens().get_text(node)


@lru_cache()
def statement_containing_node(node: ast.AST) -> ast.stmt:
    while not isinstance(node, ast.stmt):
        node = node.parent  # type: ignore
    return node


@lru_cache()
def assigned_names(node, *,
                   allow_one: bool,
                   allow_loops: bool
                   ) -> Tuple[Tuple[str, ...], ast.AST]:
    """
    Finds the names being assigned to in the nearest ancestor of
    the given node that assigns names and satisfies the given conditions.

    If allow_loops is false, this only considers assignment statements,
    e.g. `x, y = ...`. If it's true, then for loops and comprehensions are
    also considered.

    If allow_one is false, nodes which assign only one name are ignored.

    Returns:
    1. a tuple of strings containing the names of the nodes being assigned
    2. The AST node where the assignment happens
    """

    while hasattr(node, 'parent'):
        node = node.parent

        target = None

        if isinstance(node, ast.Assign):
            target = only(node.targets)
        elif isinstance(node, ast.AnnAssign):
            target = node.target
        elif isinstance(node, (ast.For, ast.comprehension)) and allow_loops:
            target = node.target

        if not target:
            continue

        names = node_names(target)
        if len(names) > 1 or allow_one:
            break
    else:
        raise TypeError('No assignment found')

    return names, node


def node_names(node: ast.AST) -> Tuple[str, ...]:
    """
    Returns a tuple of strings containing the names of
    the nodes under the given node.

    The node must be a tuple or list literal, or a single named node.

    See the doc of the function node_name.
    """
    if isinstance(node, (ast.Tuple, ast.List)):
        names = tuple(node_name(x) for x in node.elts)
    else:
        names = (node_name(node),)
    return names


def node_name(node: ast.AST) -> str:
    """
    Returns the 'name' of a node, which is either:
     - the name of a variable
     - the name of an attribute
     - the contents of a string literal key, as in d['key']
    """
    if isinstance(node, ast.Name):
        return node.id
    elif isinstance(node, ast.Attribute):
        return node.attr
    elif (isinstance(node, ast.Subscript) and
          isinstance(node.slice, ast.Index) and
          isinstance(node.slice.value, ast.Str)):  # type: ignore
        return node.slice.value.s  # type: ignore
    else:
        raise TypeError('Cannot extract name from %s' % node)


class Spell(object):
    """
    A Spell is a special callable that has information about where it's being
    called from.

    To create a spell, decorate a function with @spell.
    An instance of FrameInfo will be passed to the first argument of the function,
    while the other arguments will come from the call. For example:

        @spell
        def my_spell(frame_info, foo):
            ...

    will be called as just `my_spell(foo)`.
    """

    _excluded_codes: Set[CodeType] = set()

    # Called when decorating a function
    def __init__(self, func):
        self.func = func

    # Called when a spell is accessed as an attribute
    # (see the descriptor protocol)
    def __get__(self, instance, owner):
        # Functions are descriptors, which allow methods to
        # automatically bind the self argument.
        # Here we have to manually invoke that.
        method = self.func.__get__(instance, owner)
        return Spell(method)

    def at(self, frame_info: FrameInfo):
        """
        Returns a callable that has frame_info already bound as the first argument
        of the spell, and will accept only the other arguments normally.

        Use this to use one spell inside another.
        """
        return partial(self.func, frame_info)

    # Called when the spell is called 'plainly', e.g. my_spell(foo),
    # i.e. just as a variable without being an attribute of anything.
    # Calls where the spell is an attribute go throuh __get__.
    def __call__(self, *args, **kwargs):
        frame = sys._getframe(1)
        lookback_frames = kwargs.pop('lookback_frames', 0)

        def _lookback_until_not_excluded(frm):
            while frm.f_code in self._excluded_codes:
                frm = frm.f_back
            return frm

        frame = _lookback_until_not_excluded(frame)
        for _ in range(lookback_frames):
            frame = _lookback_until_not_excluded(frame.f_back)

        executing = Source.executing(frame)
        assert executing.node, "Failed to find call node"
        return self.at(FrameInfo(executing))(*args, **kwargs)


def __repr__(self):
        return '%s(%r)' % (
            self.__class__.__name__,
            self.func
        )


spell = Spell


def no_spells(func):
    """
    Decorate a function with this to indicate that no spells are used
    directly in this function, but the function may be used to
    access a spell dynamically. Spells looking for where they are being
    called from will skip the decorated function and look at the enclosing
    frame instead.

    For example, suppose you have a class with a method that is a spell,
    e.g.:

        class A:
            @ magic_kwargs  # makes foo a spell
            def foo(self, **kwargs):
                pass

    And another class that wraps the first:

        class B:
            def __init__(self):
                self.a = A()

    And you want users to call foo without going through A, you could write:

            @no_spells
            def foo(self, **kwargs):
                self.a.foo(**kwargs)

    Note that the method B.foo must have the same name (foo) as the spell A.foo.

    Or, if you wanted to delegate all unknown attributes to `self.a`, you could write:

            @no_spells
            def __getattr__(self, item):
                return getattr(self.a, item)

    In either case `B().foo(...)` will work as expected.
    """

    Spell._excluded_codes.add(func.__code__)
    return func



# -*- coding: future_annotations -*-
import ast
from contextlib import contextmanager
import logging
from typing import cast, TYPE_CHECKING
import sys

from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
from nbsafety.utils import fast

if TYPE_CHECKING:
    from typing import Dict, List, Optional, Set, Union


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


class AstEavesdropper(ast.NodeTransformer):
    def __init__(self, orig_to_copy_mapping: Dict[int, ast.AST]):
        self._orig_to_copy_mapping = orig_to_copy_mapping
        self._top_level_node_for_symbol: Optional[ast.AST] = None

    def _emitter_ast(self):
        return fast.Name(EMIT_EVENT, ast.Load())

    def _get_copy_id_ast(self, orig_node_id: Union[int, ast.AST]):
        if not isinstance(orig_node_id, int):
            orig_node_id = id(orig_node_id)
        return fast.Num(id(self._orig_to_copy_mapping[orig_node_id]))

    def _make_tuple_event_for(self, node: ast.AST, event: TraceEvent, orig_node_id=None, **kwargs):
        with fast.location_of(node):
            tuple_node = fast.Tuple([fast.Call(
                func=self._emitter_ast(),
                args=[event.to_ast(), self._get_copy_id_ast(orig_node_id or node)],
                keywords=[] if len(kwargs) == 0 else fast.kwargs(**kwargs),
            ), node], ast.Load())
            slc: Union[ast.Constant, ast.Num, ast.Index] = fast.Num(1)
            if sys.version_info < (3, 9):
                slc = fast.Index(slc)
            return fast.Subscript(tuple_node, slc, ast.Load())

    def visit(self, node: ast.AST):
        ret = super().visit(node)
        if isinstance(node, ast.stmt):
            # we haven't inserted statements yet, and StatementInserter needs the previous ids to be identical
            assert ret is node
        return ret

    @contextmanager
    def attrsub_load_context(self, top_level_node: Optional[ast.AST]):
        old = self._top_level_node_for_symbol
        if old is None or top_level_node is None:
            # entering context when we are already inside chain is a no-op,
            # but we can specify a context of not being in chain if we are
            # inside one (in order to support arguments)
            self._top_level_node_for_symbol = top_level_node
        yield
        self._top_level_node_for_symbol = old

    @property
    def _inside_attrsub_load_chain(self):
        return self._top_level_node_for_symbol is not None

    def visit_Attribute(self, node: ast.Attribute, call_context=False):
        with fast.location_of(node.value):
            attr_node = cast(ast.Attribute, node)
            attr_or_sub = fast.Str(attr_node.attr)
        return self.visit_Attribute_or_Subscript(node, attr_or_sub, call_context=call_context)

    def visit_Subscript(self, node: ast.Subscript, call_context=False):
        with fast.location_of(node.value):
            # TODO: expand beyond simple slices
            if isinstance(node.slice, ast.Index):
                attr_or_sub = node.slice.value  # type: ignore
                # ast.copy_location(attr_or_sub, sub_node.slice.value)
                # if isinstance(attr_or_sub, ast.Str):
                #     attr_or_sub = attr_or_sub.s
                # elif isinstance(attr_or_sub, ast.Num):
                #     attr_or_sub = attr_or_sub.n
                # else:
                #     logger.debug('unimpled index: %s', attr_or_sub)
                #     return node
            elif isinstance(node.slice, ast.Constant):
                # Python > 3.8 doesn't use ast.Index for constant slices
                attr_or_sub = node.slice
            else:
                logger.debug('unimpled slice: %s', node.slice)
                return node
            # elif isinstance(sub_node.slice, ast.Slice):
            #     raise ValueError('unimpled slice: %s' % sub_node.slice)
            # elif isinstance(sub_node.slice, ast.ExtSlice):
            #     raise ValueError('unimpled slice: %s' % sub_node.slice)
            # else:
            #     raise ValueError('unexpected slice: %s' % sub_node.slice)
        return self.visit_Attribute_or_Subscript(node, attr_or_sub, call_context=call_context)

    def _maybe_wrap_symbol_in_before_after_tracing(
            self, node, call_context=False, orig_node_id=None, begin_kwargs=None, end_kwargs=None
    ):
        if self._inside_attrsub_load_chain:
            return node
        orig_node = node
        orig_node_id = orig_node_id or id(orig_node)
        begin_kwargs = begin_kwargs or {}
        end_kwargs = end_kwargs or {}

        ctx = getattr(orig_node, 'ctx', ast.Load())
        is_load = isinstance(ctx, ast.Load)

        with fast.location_of(node):
            begin_kwargs['ret'] = self._get_copy_id_ast(orig_node_id)
            end_ret = fast.Constant(None)
            if is_load:
                end_ret = orig_node
            elif isinstance(orig_node, (ast.Attribute, ast.Subscript)):
                end_ret = orig_node.value
            end_kwargs['ret'] = end_ret
            end_kwargs['ctx'] = fast.Str(ctx.__class__.__name__)
            end_kwargs['call_context'] = fast.NameConstant(call_context)
            node = fast.Call(
                func=self._emitter_ast(),
                args=[
                    TraceEvent.after_complex_symbol.to_ast(),
                    fast.Call(
                        # this will return the node id
                        func=self._emitter_ast(),
                        args=[TraceEvent.before_complex_symbol.to_ast(), self._get_copy_id_ast(orig_node_id)],
                        keywords=fast.kwargs(**begin_kwargs),
                    )
                ],
                keywords=fast.kwargs(**end_kwargs),
            )
            if not is_load:
                if isinstance(orig_node, ast.Name):
                    node = fast.Attribute(
                        value=node,
                        attr=orig_node.id,
                    )
                elif isinstance(orig_node, ast.Attribute):
                    node = fast.Attribute(
                        value=node,
                        attr=orig_node.attr,
                    )
                elif isinstance(orig_node, ast.Subscript):
                    node = fast.Subscript(
                        value=node,
                        slice=orig_node.slice,
                    )
                else:
                    logger.error(
                        'Symbol tracing stores unsupported for node %s with type %s', orig_node, type(orig_node)
                    )
                    return orig_node
                node.ctx = ast.Store()
        # end location_of(node)
        return node

    def visit_Attribute_or_Subscript(
        self,
        node: Union[ast.Attribute, ast.Subscript],
        attr_or_sub: ast.expr,
        call_context: bool = False
    ):
        orig_node_id = id(node)
        with fast.location_of(node.value):
            extra_args: List[ast.keyword] = []
            if isinstance(node.value, ast.Name):
                extra_args = fast.kwargs(obj_name=fast.Str(node.value.id))

            with self.attrsub_load_context(node):
                node.value = fast.Call(
                    func=self._emitter_ast(),
                    args=[
                        TraceEvent.subscript.to_ast() if isinstance(node, ast.Subscript) else TraceEvent.attribute.to_ast(),
                        self._get_copy_id_ast(node.value)
                    ],
                    keywords=fast.kwargs(
                        ret=self.visit(node.value),
                        attr_or_subscript=attr_or_sub,
                        ctx=fast.Str(node.ctx.__class__.__name__),
                        call_context=fast.NameConstant(call_context),
                        top_level_node_id=self._get_copy_id_ast(self._top_level_node_for_symbol)
                    ) + extra_args
                )
        # end fast.location_of(node.value)

        return self._maybe_wrap_symbol_in_before_after_tracing(node, orig_node_id=orig_node_id)

    def _get_replacement_args(self, args, keywords: bool):
        replacement_args = []
        for arg in args:
            if keywords:
                maybe_kwarg = getattr(arg, 'value')
            else:
                maybe_kwarg = arg
            with fast.location_of(maybe_kwarg):
                with self.attrsub_load_context(None):
                    visited_maybe_kwarg = self.visit(maybe_kwarg)
                with self.attrsub_load_context(None):
                    new_arg_value = cast(ast.expr, fast.Call(
                        func=self._emitter_ast(),
                        args=[TraceEvent.argument.to_ast(), self._get_copy_id_ast(maybe_kwarg)],
                        keywords=fast.kwargs(ret=visited_maybe_kwarg),
                    ))
            if keywords:
                setattr(arg, 'value', new_arg_value)
            else:
                arg = new_arg_value
            replacement_args.append(arg)
        return replacement_args

    def visit_Call(self, node: ast.Call):
        orig_node_id = id(node)
        orig_node_func_id = id(node.func)

        with self.attrsub_load_context(node):
            if isinstance(node.func, ast.Attribute):
                node.func = self.visit_Attribute(node.func, call_context=True)
            elif isinstance(node.func, ast.Subscript):
                node.func = self.visit_Subscript(node.func, call_context=True)
            else:
                node.func = self.visit(node.func)

        # TODO: need a way to rewrite ast of subscript args,
        #  and to process these separately from outer rewrite

        node.args = self._get_replacement_args(node.args, False)
        node.keywords = self._get_replacement_args(node.keywords, True)

        # in order to ensure that the args are processed with appropriate active scope,
        # we need to make sure not to use the active namespace scope on args (in the case
        # of a function call on an ast.Attribute).
        #
        # We do so by emitting an "enter argument list", whose handler pushes the current active
        # scope while we process each argument. The "end argument list" event will then restore
        # the active scope.
        #
        # This effectively rewrites function calls as follows:
        # f(a, b, ..., c) -> trace(f, 'enter argument list')(a, b, ..., c)
        with fast.location_of(node.func):
            node.func = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.before_arg_list.to_ast(), self._get_copy_id_ast(orig_node_func_id)],
                keywords=fast.kwargs(ret=node.func),
            )

        # f(a, b, ..., c) -> trace(f(a, b, ..., c), 'exit argument list')
        with fast.location_of(node):
            node = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_arg_list.to_ast(), self._get_copy_id_ast(node)],
                keywords=fast.kwargs(ret=node),
            )

        return self._maybe_wrap_symbol_in_before_after_tracing(node, call_context=True, orig_node_id=orig_node_id)

    def visit_Assign(self, node: ast.Assign):
        if not isinstance(node.value, (ast.List, ast.Tuple)):
            return self.generic_visit(node)

        new_targets = []
        for target in node.targets:
            new_targets.append(self.visit(target))
        node.targets = cast('List[ast.expr]', new_targets)
        with fast.location_of(node.value):
            subscripted_node_value = self._make_tuple_event_for(node.value, TraceEvent.before_literal)
            node.value = fast.Call(
                func=self._emitter_ast(),
                args=[TraceEvent.after_literal.to_ast(), self._get_copy_id_ast(node.value)],
                keywords=fast.kwargs(ret=subscripted_node_value),
            )
        return node



# -*- coding: future_annotations -*-
import ast
import builtins
from collections import defaultdict
from contextlib import contextmanager
import logging
import sys
from typing import cast, TYPE_CHECKING

import astunparse

from nbsafety.analysis.attr_symbols import AttrSubSymbolChain, GetAttrSubSymbols
from nbsafety.data_model.data_symbol import DataSymbol, DataSymbolType
from nbsafety.data_model.scope import NamespaceScope
from nbsafety import singletons
from nbsafety.run_mode import SafetyRunMode
from nbsafety.singletons import nbs
from nbsafety.tracing.mutation_event import MutationEvent
from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
from nbsafety.tracing.trace_stack import TraceStack
from nbsafety.tracing.trace_stmt import TraceStatement

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Union
    from types import FrameType
    SymbolRef = Union[str, AttrSubSymbolChain]
    AttrSubVal = Union[str, int]
    RecordedArg = Tuple[AttrSubSymbolChain, int]
    RecordedArgs = Set[RecordedArg]
    DeepRef = Tuple[int, Optional[str], Tuple[RecordedArg, ...]]
    Mutation = Tuple[int, Tuple[RecordedArg, ...], MutationEvent]
    RefCandidate = Optional[Tuple[int, int, Optional[str]]]
    DeepRefCandidate = Tuple[RefCandidate, MutationEvent, RecordedArgs]
    SavedStoreData = Tuple[NamespaceScope, Any, AttrSubVal, bool]


logger = logging.getLogger(__name__)
logger.setLevel(logging.WARNING)


ARG_MUTATION_EXCEPTED_MODULES = {
    'alt',
    'altair',
    'display',
    'logging',
    'matplotlib',
    'pyplot',
    'plot',
    'plt',
    'seaborn',
    'sns',
    'widget',
}


class ListLiteral(list):
    pass


class DictLiteral(dict):
    pass


def _make_weakrefable_literal(literal):
    """
    Python dict / list / tuple can't be used in weakrefs,
    but we can force it for dict / list (but not tuple
    unfortunately) by wrapping them.
    """
    if type(literal) == list:
        return ListLiteral(literal)
    elif type(literal) == dict:
        return DictLiteral(literal)
    else:
        return literal


def _finish_tracing_reset():
    # do nothing; we just want to trigger the newly reenabled tracer with a 'call' event
    pass


class BaseTraceManager(singletons.TraceManager):

    EVENT_HANDLERS: Dict[TraceEvent, List[Callable[..., Any]]] = defaultdict(list)

    def __init__(self):
        super().__init__()
        self.tracing_enabled = False
        self.tracing_reset_pending = False

    def _emit_event(self, evt: Union[TraceEvent, str], node_id: int, **kwargs: Any):
        event = TraceEvent(evt) if isinstance(evt, str) else evt
        frame = kwargs.get('_frame', sys._getframe().f_back)
        kwargs['_frame'] = frame
        for handler in self.EVENT_HANDLERS[event]:  # type: ignore
            try:
                new_ret = handler(self, kwargs.get('ret', None), node_id, frame, event, **kwargs)
            except Exception as exc:
                if SafetyRunMode.get() == SafetyRunMode.DEVELOP:
                    raise exc
                else:
                    logger.warning('Exception occurred: %s' % exc)
                new_ret = None
            if new_ret is not None:
                kwargs['ret'] = new_ret
        return kwargs.get('ret', None)

    def _make_stack(self):
        return TraceStack(self)

    def _enable_tracing(self):
        assert not self.tracing_enabled
        self.tracing_enabled = True
        sys.settrace(self._sys_tracer)

    def _disable_tracing(self, check_enabled=True):
        if check_enabled:
            assert self.tracing_enabled
        self.tracing_enabled = False
        sys.settrace(None)

    @contextmanager
    def tracing_context(self):
        try:
            setattr(builtins, EMIT_EVENT, self._emit_event)
            self._enable_tracing()
            yield
        finally:
            delattr(builtins, EMIT_EVENT)
            self._disable_tracing(check_enabled=False)

    def _attempt_to_reenable_tracing(self, frame: FrameType):
        return NotImplemented

    def _sys_tracer(self, frame: FrameType, evt: str, *_, **__):
        if self.tracing_reset_pending:
            assert evt == 'call', 'expected call; got event %s' % evt
            self._attempt_to_reenable_tracing(frame)
            return None
        # notebook cells have filenames that appear as '<ipython-input...>'
        if evt == 'line' or not frame.f_code.co_filename.startswith('<ipython-input'):
            return None

        return self._emit_event(evt, 0, _frame=frame)


def register_handler(event: Union[TraceEvent, Tuple[TraceEvent, ...]]):
    events = event if isinstance(event, tuple) else (event,)

    def _inner_registrar(handler):
        for evt in events:
            BaseTraceManager.EVENT_HANDLERS[evt].append(handler)
        return handler
    return _inner_registrar


class TraceManager(BaseTraceManager):
    def __init__(self):
        super().__init__()
        self.trace_event_counter = 0
        self.prev_event: Optional[TraceEvent] = None
        self.prev_trace_stmt: Optional[TraceStatement] = None
        self.seen_stmts: Set[int] = set()
        self.call_depth = 0
        self.traced_statements: Dict[int, TraceStatement] = {}

        self.call_stack: TraceStack = self._make_stack()
        with self.call_stack.register_stack_state():
            # everything here should be copyable
            self.prev_trace_stmt_in_cur_frame: Optional[TraceStatement] = None
            ############################################################
            # old state:
            self.loaded_data_symbols: Set[DataSymbol] = set()
            self.saved_store_data: List[SavedStoreData] = []
            ############################################################
            # new state:
            # just something that maps orig_node_id to data symbol?
            ############################################################
            self.deep_refs: Set[DeepRef] = set()
            self.mutations: Set[Mutation] = set()
            self.deep_ref_candidates: List[DeepRefCandidate] = []
            self.literal_namespace: Optional[NamespaceScope] = None

            with self.call_stack.needing_manual_initialization():
                self.cur_frame_original_scope = nbs().global_scope
                self.active_scope = nbs().global_scope
                self.inside_lambda = False

            self.lexical_call_stack: TraceStack = self._make_stack()
            with self.lexical_call_stack.register_stack_state():
                self.first_obj_id_in_chain: Optional[int] = None
                self.saved_complex_symbol_load_info: Optional[Tuple[NamespaceScope, Tuple[Union[str, int], bool]]] = None

    def _handle_call_transition(self, trace_stmt: TraceStatement):
        new_scope = trace_stmt.get_post_call_scope()
        with self.call_stack.push():
            # TODO: figure out a better way to determine if we're inside a lambda
            #  could this one lead to a false negative if a lambda is in the default of a function def kwarg?
            self.inside_lambda = not isinstance(
                trace_stmt.stmt_node, (ast.FunctionDef, ast.AsyncFunctionDef, ast.ClassDef)
            )
            self.cur_frame_original_scope = new_scope
            self.active_scope = new_scope
        self.prev_trace_stmt_in_cur_frame = self.prev_trace_stmt = trace_stmt

    def _check_prev_stmt_done_executing_hook(self, event: TraceEvent, trace_stmt: TraceStatement):
        if event == TraceEvent.after_stmt:
            trace_stmt.finished_execution_hook()
        elif event == TraceEvent.return_ and self.prev_event not in (TraceEvent.call, TraceEvent.exception):
            # ensuring prev != call ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            if self.prev_trace_stmt is not None:
                self.prev_trace_stmt.finished_execution_hook()
            # prev_overall = self.prev_trace_stmt
            # if prev_overall is not None and prev_overall is not self._stack[-1][0]:
            #     # this condition ensures we're not inside of a stmt with multiple calls (such as map w/ lambda)
            #     prev_overall.finished_execution_hook()

    def _handle_return_transition(self, trace_stmt: TraceStatement):
        inside_lambda = self.inside_lambda
        cur_frame_scope = self.cur_frame_original_scope
        self.call_stack.pop()
        return_to_stmt = self.prev_trace_stmt_in_cur_frame
        assert return_to_stmt is not None
        if self.prev_event != TraceEvent.exception:
            # exception events are followed by return events until we hit an except clause
            # no need to track dependencies in this case
            if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
                return_to_stmt.class_scope = cast(NamespaceScope, cur_frame_scope)
            elif isinstance(trace_stmt.stmt_node, ast.Return) or inside_lambda:
                if not trace_stmt.lambda_call_point_deps_done_once:
                    trace_stmt.lambda_call_point_deps_done_once = True
                    return_to_stmt.call_point_deps.append(trace_stmt.compute_rval_dependencies())

    def state_transition_hook(
        self,
        event: TraceEvent,
        trace_stmt: TraceStatement
    ):
        self.trace_event_counter += 1

        self._check_prev_stmt_done_executing_hook(event, trace_stmt)

        if event == TraceEvent.call:
            self._handle_call_transition(trace_stmt)
        if event == TraceEvent.return_:
            self._handle_return_transition(trace_stmt)
        self.prev_event = event

    def _get_namespace_for_obj(self, obj: Any, obj_name: Optional[str] = None) -> NamespaceScope:
        obj_id = id(obj)
        ns = nbs().namespaces.get(obj_id, None)
        # print('%s attrsub %s of obj %s' % (ctx, attr_or_subscript, obj))
        if ns is not None:
            return ns
        class_scope = nbs().namespaces.get(id(obj.__class__), None)
        if class_scope is not None:
            # print('found class scope %s containing %s' % (class_scope, list(class_scope.all_data_symbols_this_indentation())))
            ns = class_scope.clone(obj)
            if obj_name is not None:
                ns.scope_name = obj_name
        else:
            # print('no scope for class', obj.__class__)
            try:
                scope_name = next(iter(nbs().aliases.get(obj_id, None))).name if obj_name is None else obj_name
            except (TypeError, StopIteration):
                scope_name = '<unknown namespace>'
            ns = NamespaceScope(obj, scope_name, parent_scope=None)
        # FIXME: brittle strategy for determining parent scope of obj
        if ns.parent_scope is None:
            if (
                obj_name is not None and
                obj_name not in self.prev_trace_stmt_in_cur_frame.frame.f_locals
            ):
                parent_scope = nbs().global_scope
            else:
                parent_scope = self.active_scope
            ns.parent_scope = parent_scope
        return ns

    def _clear_info_and_maybe_lookup_or_create_complex_symbol(self, obj_attr_or_sub) -> Optional[DataSymbol]:
        if self.saved_complex_symbol_load_info is None:
            return None
        scope, (attr_or_subscript, is_subscript) = self.saved_complex_symbol_load_info
        self.saved_complex_symbol_load_info = None
        data_sym = scope.lookup_data_symbol_by_name_this_indentation(
            attr_or_subscript, is_subscript=is_subscript
        )
        if data_sym is None:
            symbol_type = DataSymbolType.SUBSCRIPT if is_subscript else DataSymbolType.DEFAULT
            data_sym = DataSymbol(
                attr_or_subscript,
                symbol_type,
                obj_attr_or_sub,
                scope,
                stmt_node=None,
                parents=None,
                refresh_cached_obj=True,
                implicit=True,
            )
            # this is to prevent refs to the scope object from being considered as stale if we just load it
            data_sym.defined_cell_num = data_sym.required_cell_num = scope.max_defined_timestamp
            scope.put(attr_or_subscript, data_sym)
        elif data_sym.obj_id != id(obj_attr_or_sub):
            data_sym.update_obj_ref(obj_attr_or_sub)
        return data_sym

    @register_handler((TraceEvent.attribute, TraceEvent.subscript))
    def attrsub_tracer(
        self,
        obj: Any,
        _node_id: int,
        _frame_: FrameType,
        event: TraceEvent,
        *_,
        attr_or_subscript,
        ctx: str,
        call_context: bool,
        obj_name: Optional[str] = None,
        **__
    ):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        if obj is None:
            return
        self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
        is_subscript = (event == TraceEvent.subscript)
        obj_id = id(obj)
        if self.first_obj_id_in_chain is None:
            self.first_obj_id_in_chain = obj_id
        if isinstance(attr_or_subscript, tuple):
            if not all(isinstance(v, (str, int)) for v in attr_or_subscript):
                return
        elif not isinstance(attr_or_subscript, (str, int)):
            return

        scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
        self.active_scope = scope
        if ctx != 'Load':
            assert ctx in ('Store', 'AugStore')
            self.saved_store_data.append((scope, obj, attr_or_subscript, is_subscript))
            return
        if call_context:
            mutation_event = MutationEvent.normal
            if isinstance(obj, list) and attr_or_subscript == 'append':
                mutation_event = MutationEvent.list_append
            # save off event counter and obj_id
            # if event counter didn't change when we process the Call retval, and if the
            # retval is None, this is a likely signal that we have a mutation
            # TODO: this strategy won't work if the arguments themselves lead to traced function calls
            #  to cope, put DeepRefCandidates (or equivalent) in the lexical stack?
            self.deep_ref_candidates.append(
                ((self.trace_event_counter, obj_id, obj_name), mutation_event, set())
            )
        else:
            self.saved_complex_symbol_load_info = (scope, (attr_or_subscript, is_subscript))

    @register_handler(TraceEvent.after_complex_symbol)
    def after_complex_symbol(self, obj: Any, *_, call_context: bool, **__):
        try:
            if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
                return
            if self.first_obj_id_in_chain is None:
                return
            loaded_sym = self._clear_info_and_maybe_lookup_or_create_complex_symbol(obj)
            if loaded_sym is not None:
                self.loaded_data_symbols.add(loaded_sym)
            if call_context and len(self.deep_ref_candidates) > 0:
                (evt_counter, obj_id, obj_name), mutation_event, recorded_args = self.deep_ref_candidates.pop()
                if evt_counter == self.trace_event_counter:
                    if obj is None:
                        if mutation_event == MutationEvent.normal:
                            try:
                                top_level_sym = next(iter(nbs().aliases[self.first_obj_id_in_chain]))
                                if top_level_sym.is_import and top_level_sym.name not in ARG_MUTATION_EXCEPTED_MODULES:
                                    # TODO: should it be the other way around? i.e. allow-list for arg mutations, starting
                                    #  with np.random.seed?
                                    for recorded_arg, _recorded_arg_id in recorded_args:
                                        if len(recorded_arg.symbols) > 0:
                                            # only make this an arg mutation event if it looks like there's an arg to mutate
                                            mutation_event = MutationEvent.arg_mutate
                                            break
                            except:
                                pass
                        self.mutations.add((obj_id, tuple(recorded_args), mutation_event))
                    else:
                        self.deep_refs.add((obj_id, obj_name, tuple(recorded_args)))
        finally:
            self.saved_complex_symbol_load_info = None  # can't hurt to do it again
            self.first_obj_id_in_chain = None
            self.active_scope = self.cur_frame_original_scope

    @register_handler(TraceEvent.argument)
    def arg_recorder(self, arg_obj: Any, arg_node_id: int, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        arg_node = nbs().ast_node_by_id.get(arg_node_id, None)
        if not isinstance(arg_node, (ast.Attribute, ast.Subscript, ast.Call, ast.Name)):
            return
        if len(self.deep_ref_candidates) == 0:
            return

        arg_obj_id = id(arg_obj)
        # TODO: we should be able to get the actual data symbol during live tracing,
        #  instead of trying to resolve from an attrsub chain determined via analysis
        recorded_arg = GetAttrSubSymbols()(arg_node)
        self.deep_ref_candidates[-1][-1].add((recorded_arg, arg_obj_id))

    @register_handler(TraceEvent.before_arg_list)
    def before_argument_list(self, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        with self.lexical_call_stack.push():
            pass
        self.active_scope = self.cur_frame_original_scope

    @register_handler(TraceEvent.after_arg_list)
    def after_argument_list(self, *_, **__):
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return
        # no need to reset active scope here;
        # that will happen in the 'after chain' handler
        self.lexical_call_stack.pop()

    @register_handler(TraceEvent.after_literal)
    def after_literal(self, literal: Any, *_, **__):
        literal = _make_weakrefable_literal(literal)
        if not self.tracing_enabled or self.prev_trace_stmt_in_cur_frame.finished:
            return literal
        if self.prev_trace_stmt_in_cur_frame.finished:
            return literal
        if isinstance(literal, (dict, list, tuple)):
            scope = NamespaceScope(
                literal, None, self.prev_trace_stmt_in_cur_frame.scope
            )
            gen = literal.items() if isinstance(literal, dict) else enumerate(literal)
            for i, obj in gen:
                scope.upsert_data_symbol_for_name(
                    i, obj, set(), self.prev_trace_stmt_in_cur_frame.stmt_node, True
                )
            self.literal_namespace = scope
        return literal

    @register_handler(TraceEvent.after_stmt)
    def after_stmt(self, ret_expr: Any, stmt_id: int, frame: FrameType, *_, **__):
        if stmt_id in self.seen_stmts:
            return ret_expr
        stmt = nbs().ast_node_by_id.get(stmt_id, None)
        if stmt is not None:
            self.handle_sys_events(None, 0, frame, TraceEvent.after_stmt, stmt_node=cast(ast.stmt, stmt))
        return ret_expr

    @register_handler(TraceEvent.before_stmt)
    def before_stmt(self, _ret: None, stmt_id: int, frame: FrameType, *_, **__) -> None:
        if stmt_id in self.seen_stmts:
            return
        # logger.warning('reenable tracing: %s', site_id)
        if self.prev_trace_stmt_in_cur_frame is not None:
            prev_trace_stmt_in_cur_frame = self.prev_trace_stmt_in_cur_frame
            # both of the following stmts should be processed when body is entered
            if isinstance(prev_trace_stmt_in_cur_frame.stmt_node, (ast.For, ast.If, ast.With)):
                self.after_stmt(None, prev_trace_stmt_in_cur_frame.stmt_id, frame)
        trace_stmt = self.traced_statements.get(stmt_id, None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(
                frame,
                cast(ast.stmt, nbs().ast_node_by_id[stmt_id]),
                self.cur_frame_original_scope
            )
            self.traced_statements[stmt_id] = trace_stmt
        self.prev_trace_stmt_in_cur_frame = trace_stmt
        self.prev_trace_stmt = trace_stmt
        if not self.tracing_enabled:
            assert not self.tracing_reset_pending
            self._enable_tracing()
            self.tracing_reset_pending = True
            _finish_tracing_reset()  # trigger the tracer with a frame

    def after_stmt_reset_hook(self):
        self.loaded_data_symbols.clear()
        self.saved_store_data.clear()
        self.deep_refs.clear()
        self.mutations.clear()
        self.deep_ref_candidates.clear()
        self.lexical_call_stack.clear()
        self.active_scope = self.cur_frame_original_scope
        self.literal_namespace = None
        self.first_obj_id_in_chain = None

    def _attempt_to_reenable_tracing(self, frame: FrameType) -> None:
        if nbs().is_develop:
            assert self.tracing_reset_pending, 'expected tracing reset to be pending!'
            assert self.call_depth > 0, 'expected managed call depth > 0, got %d' % self.call_depth
        self.tracing_reset_pending = False
        call_depth = 0
        while frame is not None:
            if frame.f_code.co_filename.startswith('<ipython-input'):
                call_depth += 1
            frame = frame.f_back
        if nbs().is_develop:
            assert call_depth >= 1, 'expected call depth >= 1, got %d' % call_depth
        # TODO: allow reenabling tracing beyond just at the top level
        if call_depth != 1:
            self._disable_tracing()
            return
        # at this point, we can be sure we're at the top level
        # because tracing was enabled in a handler and not in the
        # top level, we need to clear the stack, since we won't
        # catch the return event
        self.call_depth = 0
        self.call_stack.clear()
        self.lexical_call_stack.clear()
        if nbs().settings.trace_messages_enabled:
            logger.warning('reenable tracing >>>')

    @register_handler((TraceEvent.call, TraceEvent.return_, TraceEvent.exception))
    def handle_sys_events(
        self,
        _ret: None,
        _node_id: int,
        frame: FrameType,
        event: TraceEvent,
        *_,
        stmt_node: Optional[ast.stmt] = None,
        **__
    ):
        # right now, this should only be enabled for notebook code
        assert frame.f_code.co_filename.startswith('<ipython-input')
        assert self.tracing_enabled or event == TraceEvent.after_stmt
        nbs().maybe_set_name_to_cell_num_mapping(frame)

        # IPython quirk -- every line in outer scope apparently wrapped in lambda
        # We want to skip the outer 'call' and 'return' for these
        if event == TraceEvent.call:
            self.call_depth += 1
            if self.call_depth == 1:
                return self._sys_tracer

        if event == TraceEvent.return_:
            self.call_depth -= 1
            if self.call_depth == 0:
                return self._sys_tracer

        cell_num, lineno = nbs().get_position(frame)

        if event == TraceEvent.after_stmt:
            assert stmt_node is not None
        else:
            try:
                stmt_node = nbs().statement_cache[cell_num][lineno]
            except KeyError as e:
                logger.warning("got key error for stmt node in cell %d, line %d", cell_num, lineno)
                if nbs().is_develop:
                    raise e
                return self._sys_tracer

        trace_stmt = self.traced_statements.get(id(stmt_node), None)
        if trace_stmt is None:
            trace_stmt = TraceStatement(frame, stmt_node, self.cur_frame_original_scope)
            self.traced_statements[id(stmt_node)] = trace_stmt

        if nbs().settings.trace_messages_enabled:
            codeline = astunparse.unparse(stmt_node).strip('\n').split('\n')[0]
            codeline = ' ' * getattr(stmt_node, 'col_offset', 0) + codeline
            logger.warning(' %3d: %10s >>> %s', lineno, event, codeline)
        if event == TraceEvent.call:
            if trace_stmt.call_seen:
                if nbs().settings.trace_messages_enabled:
                    logger.warning(' disable tracing >>>')
                self._disable_tracing()
                return None
            trace_stmt.call_seen = True
        self.state_transition_hook(event, trace_stmt)
        return self._sys_tracer



from .core import spell, no_spells

from .spells import unpack_keys, unpack_attrs, args_with_source, dict_of, print_args, call_with_name, delegate_to_attr, \
    maybe, select_from, magic_kwargs, assigned_names, switch, timeit

__version__ = '0.2.1'



# -*- coding: future_annotations -*-
from .safety_ast_rewriter import SafetyAstRewriter
from .trace_manager import TraceManager
from .trace_events import TraceEvent
from .trace_stmt import TraceStatement



# -*- coding: future_annotations -*-
# import gc
# import weakref
# import sys


class CommonEqualityMixin:
    def __eq__(self, other):
        return isinstance(other, self.__class__) and (self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not (self == other)


# class EnforceSingletonMixin:
#     _INSTANCE = None
#
#     def __init__(self):
#         if self.__class__._INSTANCE is not None and self.__class__._INSTANCE() is not None:
#             gc.collect()
#             print(sys.getrefcount(self.__class__._INSTANCE()))
#             print(self.__class__._INSTANCE())
#             print(gc.get_referrers(self.__class__._INSTANCE()))
#             assert False
#         self.__class__._INSTANCE = weakref.ref(self)
#
#     def __del__(self):
#         self.__class__._INSTANCE = None
